from pixel_font_builder import Glyph


def test_glyph_1():
    glyph = Glyph(
        name='test',
        horizontal_offset=(1, 2),
        advance_width=3,
        vertical_offset=(4, 5),
        advance_height=6,
        bitmap=[
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 1, 0, 0],
            [0, 1, 1, 0, 0, 0, 0],
            [0, 0, 1, 1, 1, 0, 0],
            [0, 0, 1, 0, 1, 0, 0],
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
        ],
    )
    assert glyph.name == 'test'
    assert glyph.horizontal_offset_x == 1
    assert glyph.horizontal_offset_y == 2
    assert glyph.horizontal_offset == (1, 2)
    assert glyph.advance_width == 3
    assert glyph.vertical_offset_x == 4
    assert glyph.vertical_offset_y == 5
    assert glyph.vertical_offset == (4, 5)
    assert glyph.advance_height == 6
    assert glyph.width == 7
    assert glyph.height == 9
    assert glyph.dimensions == (7, 9)
    assert glyph.calculate_bitmap_left_padding() == 1
    assert glyph.calculate_bitmap_right_padding() == 2
    assert glyph.calculate_bitmap_top_padding() == 3
    assert glyph.calculate_bitmap_bottom_padding() == 1

    glyph.horizontal_offset = (7, 8)
    assert glyph.horizontal_offset_x == 7
    assert glyph.horizontal_offset_y == 8

    glyph.vertical_offset = (9, 10)
    assert glyph.vertical_offset_x == 9
    assert glyph.vertical_offset_y == 10


def test_glyph_2():
    glyph = Glyph(name='test')
    assert glyph.horizontal_offset == (0, 0)
    assert glyph.advance_width == 0
    assert glyph.vertical_offset == (0, 0)
    assert glyph.advance_height == 0
    assert glyph.dimensions == (0, 0)
    assert glyph.calculate_bitmap_left_padding() == 0
    assert glyph.calculate_bitmap_right_padding() == 0
    assert glyph.calculate_bitmap_top_padding() == 0
    assert glyph.calculate_bitmap_bottom_padding() == 0


def test_glyph_3():
    glyph = Glyph(
        name='test',
        bitmap=[
            [0, 0],
            [0, 0],
            [0, 0],
            [0, 0],
        ],
    )
    assert glyph.calculate_bitmap_left_padding() == 2
    assert glyph.calculate_bitmap_right_padding() == 2
    assert glyph.calculate_bitmap_top_padding() == 4
    assert glyph.calculate_bitmap_bottom_padding() == 4



from pixel_font_builder.meta import WeightName, MetaInfo


def create_name_strings(meta_info: MetaInfo) -> dict[str, str]:
    """
    https://learn.microsoft.com/en-us/typography/opentype/spec/name#name-ids
    copyright (nameID 0)
    familyName (nameID 1)
    styleName (nameID 2)
    uniqueFontIdentifier (nameID 3)
    fullName (nameID 4)
    version (nameID 5)
    psName (nameID 6)
    trademark (nameID 7)
    manufacturer (nameID 8)
    designer (nameID 9)
    description (nameID 10)
    vendorURL (nameID 11)
    designerURL (nameID 12)
    licenseDescription (nameID 13)
    licenseInfoURL (nameID 14)
    typographicFamily (nameID 16)
    typographicSubfamily (nameID 17)
    compatibleFullName (nameID 18)
    sampleText (nameID 19)
    postScriptCIDFindfontName (nameID 20)
    wwsFamilyName (nameID 21)
    wwsSubfamilyName (nameID 22)
    lightBackgroundPalette (nameID 23)
    darkBackgroundPalette (nameID 24)
    variationsPostScriptNamePrefix (nameID 25)
    """
    unique_name = meta_info.family_name.replace(' ', '-')
    style_name = meta_info.weight_name or WeightName.REGULAR
    name_strings = {
        'familyName': meta_info.family_name,
        'styleName': style_name,
        'uniqueFontIdentifier': f'{unique_name}-{style_name};{meta_info.version}',
        'fullName': f'{meta_info.family_name} {style_name}',
        'version': meta_info.version,
        'psName': f'{unique_name}-{style_name}',
    }
    if meta_info.copyright_info is not None:
        name_strings['copyright'] = meta_info.copyright_info
    if meta_info.manufacturer is not None:
        name_strings['manufacturer'] = meta_info.manufacturer
    if meta_info.designer is not None:
        name_strings['designer'] = meta_info.designer
    if meta_info.description is not None:
        name_strings['description'] = meta_info.description
    if meta_info.vendor_url is not None:
        name_strings['vendorURL'] = meta_info.vendor_url
    if meta_info.designer_url is not None:
        name_strings['designerURL'] = meta_info.designer_url
    if meta_info.license_info is not None:
        name_strings['licenseDescription'] = meta_info.license_info
    if meta_info.license_url is not None:
        name_strings['licenseInfoURL'] = meta_info.license_url
    if meta_info.sample_text is not None:
        name_strings['sampleText'] = meta_info.sample_text
    return name_strings



from typing import Final

from pixel_font_builder.opentype.feature import FeatureFile
from pixel_font_builder.opentype.outline import OutlinesPainter, SolidOutlinesPainter


class Config:
    DEFAULT_OUTLINES_PAINTER: Final = SolidOutlinesPainter()

    px_to_units: int
    has_vertical_metrics: bool
    outlines_painter: OutlinesPainter
    feature_files: list[FeatureFile]

    def __init__(
            self,
            px_to_units: int = 100,
            has_vertical_metrics: bool = True,
            outlines_painter: OutlinesPainter | None = None,
            feature_files: list[FeatureFile] | None = None,
    ):
        self.px_to_units = px_to_units
        self.has_vertical_metrics = has_vertical_metrics
        self.outlines_painter = Config.DEFAULT_OUTLINES_PAINTER if outlines_painter is None else outlines_painter
        self.feature_files = [] if feature_files is None else feature_files



import shutil
from datetime import datetime

from examples import build_dir
from pixel_font_builder import FontBuilder, WeightName, SerifStyle, SlantStyle, WidthStyle, Glyph, opentype


def main():
    outputs_dir = build_dir.joinpath('create')
    if outputs_dir.exists():
        shutil.rmtree(outputs_dir)
    outputs_dir.mkdir(parents=True)

    builder = FontBuilder()
    builder.font_metric.font_size = 16
    builder.font_metric.horizontal_layout.ascent = 14
    builder.font_metric.horizontal_layout.descent = -2
    builder.font_metric.vertical_layout.ascent = 8
    builder.font_metric.vertical_layout.descent = -8
    builder.font_metric.x_height = 7
    builder.font_metric.cap_height = 10
    builder.font_metric.underline_position = -2
    builder.font_metric.underline_thickness = 1
    builder.font_metric.strikeout_position = 6
    builder.font_metric.strikeout_thickness = 1

    builder.meta_info.version = '1.0.0'
    builder.meta_info.created_time = datetime.fromisoformat('2024-01-01T00:00:00Z')
    builder.meta_info.modified_time = builder.meta_info.created_time
    builder.meta_info.family_name = 'My Font'
    builder.meta_info.weight_name = WeightName.REGULAR
    builder.meta_info.serif_style = SerifStyle.SANS_SERIF
    builder.meta_info.slant_style = SlantStyle.NORMAL
    builder.meta_info.width_style = WidthStyle.MONOSPACED
    builder.meta_info.manufacturer = 'Pixel Font Studio'
    builder.meta_info.designer = 'TakWolf'
    builder.meta_info.description = 'A pixel font'
    builder.meta_info.copyright_info = 'Copyright (c) TakWolf'
    builder.meta_info.license_info = 'This Font Software is licensed under the SIL Open Font License, Version 1.1'
    builder.meta_info.vendor_url = 'https://github.com/TakWolf/pixel-font-builder'
    builder.meta_info.designer_url = 'https://takwolf.com'
    builder.meta_info.license_url = 'https://openfontlicense.org'

    builder.character_mapping.update({
        65: 'CAP_LETTER_A',
    })

    builder.glyphs.append(Glyph(
        name='.notdef',
        horizontal_offset=(0, -2),
        advance_width=8,
        vertical_offset=(-4, 0),
        advance_height=16,
        bitmap=[
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 1, 1, 1, 1, 1, 1, 1],
        ],
    ))
    builder.glyphs.append(Glyph(
        name='CAP_LETTER_A',
        horizontal_offset=(0, -2),
        advance_width=8,
        vertical_offset=(-4, 0),
        advance_height=16,
        bitmap=[
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
        ],
    ))

    builder.save_otf(outputs_dir.joinpath('my-font.otf'))
    builder.save_otf(outputs_dir.joinpath('my-font.otf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_otf(outputs_dir.joinpath('my-font.otf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(outputs_dir.joinpath('my-font.ttf'))
    builder.save_ttf(outputs_dir.joinpath('my-font.ttf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_ttf(outputs_dir.joinpath('my-font.ttf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_bdf(outputs_dir.joinpath('my-font.bdf'))
    builder.save_pcf(outputs_dir.joinpath('my-font.pcf'))

    builder.meta_info.family_name = 'My Font SquareDot'
    builder.opentype_config.outlines_painter = opentype.SquareDotOutlinesPainter()
    builder.save_otf(outputs_dir.joinpath('my-font-square_dot.otf'))
    builder.save_otf(outputs_dir.joinpath('my-font-square_dot.otf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_otf(outputs_dir.joinpath('my-font-square_dot.otf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(outputs_dir.joinpath('my-font-square_dot.ttf'))
    builder.save_ttf(outputs_dir.joinpath('my-font-square_dot.ttf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_ttf(outputs_dir.joinpath('my-font-square_dot.ttf.woff2'), flavor=opentype.Flavor.WOFF2)

    builder.meta_info.family_name = 'My Font CircleDot'
    builder.opentype_config.outlines_painter = opentype.CircleDotOutlinesPainter()
    builder.save_otf(outputs_dir.joinpath('my-font-circle_dot.otf'))
    builder.save_otf(outputs_dir.joinpath('my-font-circle_dot.otf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_otf(outputs_dir.joinpath('my-font-circle_dot.otf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(outputs_dir.joinpath('my-font-circle_dot.ttf'))
    builder.save_ttf(outputs_dir.joinpath('my-font-circle_dot.ttf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_ttf(outputs_dir.joinpath('my-font-circle_dot.ttf.woff2'), flavor=opentype.Flavor.WOFF2)


if __name__ == '__main__':
    main()



from pixel_font_builder import LineMetric, FontMetric


def test_font_metric_1():
    font_metric = FontMetric(
        font_size=10,
        horizontal_layout=LineMetric(
            ascent=1,
            descent=-1,
            line_gap=2,
        ),
        vertical_layout=LineMetric(
            ascent=3,
            descent=-3,
            line_gap=4,
        ),
        x_height=5,
        cap_height=6,
        underline_position=7,
        underline_thickness=8,
        strikeout_position=9,
        strikeout_thickness=10,
    ) * 2
    assert font_metric.font_size == 20
    assert font_metric.horizontal_layout.ascent == 2
    assert font_metric.horizontal_layout.descent == -2
    assert font_metric.horizontal_layout.line_height == 4
    assert font_metric.horizontal_layout.line_gap == 4
    assert font_metric.vertical_layout.ascent == 6
    assert font_metric.vertical_layout.descent == -6
    assert font_metric.vertical_layout.line_height == 12
    assert font_metric.vertical_layout.line_gap == 8
    assert font_metric.x_height == 10
    assert font_metric.cap_height == 12
    assert font_metric.underline_position == 14
    assert font_metric.underline_thickness == 16
    assert font_metric.strikeout_position == 18
    assert font_metric.strikeout_thickness == 20


def test_font_metric_2():
    font_metric = FontMetric()
    assert font_metric.font_size == 0
    assert font_metric.horizontal_layout.ascent == 0
    assert font_metric.horizontal_layout.descent == 0
    assert font_metric.horizontal_layout.line_height == 0
    assert font_metric.horizontal_layout.line_gap == 0
    assert font_metric.vertical_layout.ascent == 0
    assert font_metric.vertical_layout.descent == 0
    assert font_metric.vertical_layout.line_height == 0
    assert font_metric.vertical_layout.line_gap == 0
    assert font_metric.x_height == 0
    assert font_metric.cap_height == 0
    assert font_metric.underline_position == 0
    assert font_metric.underline_thickness == 0
    assert font_metric.strikeout_position == 0
    assert font_metric.strikeout_thickness == 0



from __future__ import annotations

from enum import StrEnum

from fontTools.fontBuilder import FontBuilder
from fontTools.misc import timeTools
from fontTools.ttLib import TTCollection

import pixel_font_builder
from pixel_font_builder.opentype.feature import build_kern_feature
from pixel_font_builder.opentype.name import create_name_strings
from pixel_font_builder.opentype.outline import create_xtf_glyphs


class Flavor(StrEnum):
    WOFF = 'woff'
    WOFF2 = 'woff2'


def create_font_builder(
        context: pixel_font_builder.FontBuilder,
        is_ttf: bool,
        flavor: Flavor | None = None,
) -> FontBuilder:
    config = context.opentype_config
    font_metric = context.font_metric * config.px_to_units
    meta_info = context.meta_info
    character_mapping = context.character_mapping
    glyph_order, name_to_glyph = context.prepare_glyphs()
    kerning_pairs = context.kerning_pairs

    builder = FontBuilder(font_metric.font_size, isTTF=is_ttf)

    if meta_info.created_time is not None:
        setattr(builder.font['head'], 'created', timeTools.timestampSinceEpoch(meta_info.created_time.timestamp()))
    if meta_info.modified_time is not None:
        setattr(builder.font['head'], 'modified', timeTools.timestampSinceEpoch(meta_info.modified_time.timestamp()))

    name_strings = create_name_strings(meta_info)
    builder.setupNameTable(name_strings)

    xtf_glyphs, horizontal_metrics, vertical_metrics = create_xtf_glyphs(is_ttf, config.outlines_painter, name_to_glyph, config.px_to_units)
    builder.setupGlyphOrder(glyph_order)
    builder.setupGlyf(xtf_glyphs) if is_ttf else builder.setupCFF('', {}, xtf_glyphs, {})
    builder.setupHorizontalMetrics(horizontal_metrics)
    if config.has_vertical_metrics:
        builder.setupVerticalMetrics(vertical_metrics)

    builder.setupCharacterMap(character_mapping)

    builder.setupHorizontalHeader(
        ascent=font_metric.horizontal_layout.ascent,
        descent=font_metric.horizontal_layout.descent,
        lineGap=font_metric.horizontal_layout.line_gap,
    )
    if config.has_vertical_metrics:
        builder.setupVerticalHeader(
            ascent=font_metric.vertical_layout.ascent,
            descent=font_metric.vertical_layout.descent,
            lineGap=font_metric.vertical_layout.line_gap,
        )
    builder.setupOS2(
        sTypoAscender=font_metric.horizontal_layout.ascent,
        sTypoDescender=font_metric.horizontal_layout.descent,
        sTypoLineGap=font_metric.horizontal_layout.line_gap,
        usWinAscent=max(font_metric.horizontal_layout.ascent, 0),
        usWinDescent=max(-font_metric.horizontal_layout.descent, 0),
        sxHeight=font_metric.x_height,
        sCapHeight=font_metric.cap_height,
        yStrikeoutPosition=font_metric.strikeout_position,
        yStrikeoutSize=font_metric.strikeout_thickness,
        ulCodePageRange1=0b_11100000_00111111_00000001_11111111,
        ulCodePageRange2=0b_11111111_11111111_00000000_00000000,
    )
    builder.setupPost(
        underlinePosition=font_metric.underline_position,
        underlineThickness=font_metric.underline_thickness,
    )

    if len(kerning_pairs) > 0:
        builder.addOpenTypeFeatures(build_kern_feature(kerning_pairs, config.px_to_units))

    for feature_file in config.feature_files:
        builder.addOpenTypeFeatures(feature_file.text, feature_file.file_path)

    if flavor is not None:
        builder.font.flavor = flavor

    return builder


def create_font_collection_builder(
        contexts: pixel_font_builder.FontCollectionBuilder,
        is_ttf: bool,
) -> TTCollection:
    collection_builder = TTCollection()
    collection_builder.fonts.extend(create_font_builder(context, is_ttf).font for context in contexts)
    return collection_builder



from datetime import datetime
from enum import StrEnum
from typing import Any


class WeightName(StrEnum):
    LIGHT = 'Light'
    NORMAL = 'Normal'
    REGULAR = 'Regular'
    MEDIUM = 'Medium'
    BOLD = 'Bold'
    HEAVY = 'Heavy'


class SerifStyle(StrEnum):
    SERIF = 'Serif'
    SANS_SERIF = 'Sans Serif'


class SlantStyle(StrEnum):
    NORMAL = 'Normal'
    ITALIC = 'Italic'
    OBLIQUE = 'Oblique'
    REVERSE_ITALIC = 'Reverse Italic'
    REVERSE_OBLIQUE = 'Reverse Oblique'


class WidthStyle(StrEnum):
    MONOSPACED = 'Monospaced'
    DUOSPACED = 'Duospaced'
    CHARACTER_CELL = 'Character Cell'
    PROPORTIONAL = 'Proportional'


class MetaInfo:
    version: str
    created_time: datetime | None
    modified_time: datetime | None
    family_name: str | None
    weight_name: WeightName | None
    serif_style: SerifStyle | None
    slant_style: SlantStyle | None
    width_style: WidthStyle | None
    manufacturer: str | None
    designer: str | None
    description: str | None
    copyright_info: str | None
    license_info: str | None
    vendor_url: str | None
    designer_url: str | None
    license_url: str | None
    sample_text: str | None

    def __init__(
            self,
            version: str = '0.0.0',
            created_time: datetime | None = None,
            modified_time: datetime | None = None,
            family_name: str | None = None,
            weight_name: WeightName | None = None,
            serif_style: SerifStyle | None = None,
            slant_style: SlantStyle | None = None,
            width_style: WidthStyle | None = None,
            manufacturer: str | None = None,
            designer: str | None = None,
            description: str | None = None,
            copyright_info: str | None = None,
            license_info: str | None = None,
            vendor_url: str | None = None,
            designer_url: str | None = None,
            license_url: str | None = None,
            sample_text: str | None = None,
    ):
        self.version = version
        self.created_time = created_time
        self.modified_time = modified_time
        self.family_name = family_name
        self.weight_name = weight_name
        self.serif_style = serif_style
        self.slant_style = slant_style
        self.width_style = width_style
        self.manufacturer = manufacturer
        self.designer = designer
        self.description = description
        self.copyright_info = copyright_info
        self.license_info = license_info
        self.vendor_url = vendor_url
        self.designer_url = designer_url
        self.license_url = license_url
        self.sample_text = sample_text

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MetaInfo):
            return NotImplemented
        return (self.version == other.version and
                self.created_time == other.created_time and
                self.modified_time == other.modified_time and
                self.family_name == other.family_name and
                self.weight_name == other.weight_name and
                self.serif_style == other.serif_style and
                self.slant_style == other.slant_style and
                self.width_style == other.width_style and
                self.manufacturer == other.manufacturer and
                self.designer == other.designer and
                self.description == other.description and
                self.copyright_info == other.copyright_info and
                self.license_info == other.license_info and
                self.vendor_url == other.vendor_url and
                self.designer_url == other.designer_url and
                self.license_url == other.license_url and
                self.sample_text == other.sample_text)



from typing import Any


class Glyph:
    name: str
    horizontal_offset_x: int
    horizontal_offset_y: int
    advance_width: int
    vertical_offset_x: int
    vertical_offset_y: int
    advance_height: int
    bitmap: list[list[int]]

    def __init__(
            self,
            name: str,
            horizontal_offset: tuple[int, int] = (0, 0),
            advance_width: int = 0,
            vertical_offset: tuple[int, int] = (0, 0),
            advance_height: int = 0,
            bitmap: list[list[int]] | None = None,
    ):
        self.name = name
        self.horizontal_offset_x, self.horizontal_offset_y = horizontal_offset
        self.advance_width = advance_width
        self.vertical_offset_x, self.vertical_offset_y = vertical_offset
        self.advance_height = advance_height
        self.bitmap = [] if bitmap is None else bitmap

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Glyph):
            return NotImplemented
        return (self.name == other.name and
                self.horizontal_offset_x == other.horizontal_offset_x and
                self.horizontal_offset_y == other.horizontal_offset_y and
                self.advance_width == other.advance_width and
                self.vertical_offset_x == other.vertical_offset_x and
                self.vertical_offset_y == other.vertical_offset_y and
                self.advance_height == other.advance_height and
                self.bitmap == other.bitmap)

    @property
    def horizontal_offset(self) -> tuple[int, int]:
        return self.horizontal_offset_x, self.horizontal_offset_y

    @horizontal_offset.setter
    def horizontal_offset(self, value: tuple[int, int]):
        self.horizontal_offset_x, self.horizontal_offset_y = value

    @property
    def vertical_offset(self) -> tuple[int, int]:
        return self.vertical_offset_x, self.vertical_offset_y

    @vertical_offset.setter
    def vertical_offset(self, value: tuple[int, int]):
        self.vertical_offset_x, self.vertical_offset_y = value

    @property
    def width(self) -> int:
        if len(self.bitmap) > 0:
            return len(self.bitmap[0])
        else:
            return 0

    @property
    def height(self) -> int:
        return len(self.bitmap)

    @property
    def dimensions(self) -> tuple[int, int]:
        return self.width, self.height

    def calculate_bitmap_left_padding(self) -> int:
        padding = 0
        for i in range(self.width):
            if any(bitmap_row[i] != 0 for bitmap_row in self.bitmap):
                break
            padding += 1
        return padding

    def calculate_bitmap_right_padding(self) -> int:
        padding = 0
        for i in range(self.width):
            if any(bitmap_row[-1 - i] != 0 for bitmap_row in self.bitmap):
                break
            padding += 1
        return padding

    def calculate_bitmap_top_padding(self) -> int:
        padding = 0
        for bitmap_row in self.bitmap:
            if any(color != 0 for color in bitmap_row):
                break
            padding += 1
        return padding

    def calculate_bitmap_bottom_padding(self) -> int:
        padding = 0
        for bitmap_row in reversed(self.bitmap):
            if any(color != 0 for color in bitmap_row):
                break
            padding += 1
        return padding



from pixel_font_builder.opentype.common import Flavor, create_font_builder, create_font_collection_builder
from pixel_font_builder.opentype.config import Config
from pixel_font_builder.opentype.feature import FeatureFile
from pixel_font_builder.opentype.outline import OutlinesPen, OutlinesPainter, SolidOutlinesPainter, SquareDotOutlinesPainter, CircleDotOutlinesPainter




class Config:
    resolution_x: int
    resolution_y: int
    draw_right_to_left: bool
    ms_byte_first: bool
    ms_bit_first: bool
    glyph_pad_index: int
    scan_unit_index: int

    def __init__(
            self,
            resolution_x: int = 75,
            resolution_y: int = 75,
            draw_right_to_left: bool = False,
            ms_byte_first: bool = True,
            ms_bit_first: bool = True,
            glyph_pad_index: int = 0,
            scan_unit_index: int = 0,
    ):
        self.resolution_x = resolution_x
        self.resolution_y = resolution_y
        self.draw_right_to_left = draw_right_to_left
        self.ms_byte_first = ms_byte_first
        self.ms_bit_first = ms_bit_first
        self.glyph_pad_index = glyph_pad_index
        self.scan_unit_index = scan_unit_index



from __future__ import annotations

from io import StringIO
from os import PathLike


class FeatureFile:
    @staticmethod
    def load(file_path: str | PathLike[str]) -> FeatureFile:
        with open(file_path, 'r', encoding='utf-8') as file:
            return FeatureFile(file.read(), file_path)

    text: str
    file_path: str | PathLike[str] | None

    def __init__(
            self,
            text: str,
            file_path: str | PathLike[str] | None = None,
    ):
        self.text = text
        self.file_path = file_path


def build_kern_feature(pairs: dict[tuple[str, str], int], px_to_units: int) -> str:
    text = StringIO()
    text.write('languagesystem DFLT dflt;\n')
    text.write('\n')
    text.write('feature kern {\n')
    for (left_glyph_name, right_glyph_name), offset in sorted(pairs.items()):
        text.write(f'    position {left_glyph_name} {right_glyph_name} {offset * px_to_units};\n')
    text.write('} kern;\n')
    return text.getvalue()



from __future__ import annotations

import math
import shutil
from datetime import datetime
from pathlib import Path

import png

from examples import glyphs_dir, build_dir
from pixel_font_builder import FontBuilder, FontCollectionBuilder, WeightName, SerifStyle, SlantStyle, WidthStyle, Glyph, opentype


def _load_bitmap_from_png(file_path: Path) -> tuple[list[list[int]], int, int]:
    width, height, pixels, _ = png.Reader(filename=file_path).read()
    bitmap = []
    for pixels_row in pixels:
        bitmap_row = []
        for i in range(0, width * 4, 4):
            alpha = pixels_row[i + 3]
            bitmap_row.append(1 if alpha > 127 else 0)
        bitmap.append(bitmap_row)
    return bitmap, width, height


def _save_bitmap_to_png(bitmap: list[list[int]], file_path: Path):
    pixels = []
    for bitmap_row in bitmap:
        pixels_row = []
        for color in bitmap_row:
            pixels_row.append(0)
            pixels_row.append(0)
            pixels_row.append(0)
            pixels_row.append(0 if color == 0 else 255)
        pixels.append(pixels_row)
    png.from_array(pixels, 'RGBA').save(file_path)


class GlyphFile:
    @staticmethod
    def load(file_path: Path) -> GlyphFile:
        hex_name = file_path.stem.strip()
        code_point = -1 if hex_name == 'notdef' else int(hex_name, 16)
        return GlyphFile(file_path, code_point)

    file_path: Path
    code_point: int
    bitmap: list[list[int]]
    width: int
    height: int

    def __init__(self, file_path: Path, code_point: int):
        self.file_path = file_path
        self.code_point = code_point
        self.bitmap, self.width, self.height = _load_bitmap_from_png(file_path)

    def save(self):
        _save_bitmap_to_png(self.bitmap, self.file_path)


def _get_glyph_name(code_point: int) -> str:
    return '.notdef' if code_point == -1 else f'u{code_point:04X}'


def _collect_glyph_files() -> tuple[dict[int, str], list[GlyphFile]]:
    character_mapping = {}
    glyph_files = []
    for file_path in glyphs_dir.iterdir():
        if file_path.suffix != '.png':
            continue
        glyph_file = GlyphFile.load(file_path)
        if glyph_file.code_point != -1:
            character_mapping[glyph_file.code_point] = _get_glyph_name(glyph_file.code_point)
        glyph_files.append(glyph_file)
    glyph_files.sort(key=lambda x: x.code_point)
    return character_mapping, glyph_files


def _build_kerning_pairs() -> dict[tuple[str, str], int]:
    kerning_pairs = {}
    for left_letter in 'FT':
        for right_letter in 'acdegmnopqrsuvwxyz':
            kerning_pairs[(_get_glyph_name(ord(left_letter)), _get_glyph_name(ord(right_letter)))] = -1
    return kerning_pairs


def _create_builder(
        character_mapping: dict[int, str],
        glyph_files: list[GlyphFile],
        kerning_pairs: dict[tuple[str, str], int],
        name_num: int = 0,
) -> FontBuilder:
    builder = FontBuilder()
    builder.font_metric.font_size = 11
    builder.font_metric.horizontal_layout.ascent = 11
    builder.font_metric.horizontal_layout.descent = -4
    builder.font_metric.vertical_layout.ascent = 8
    builder.font_metric.vertical_layout.descent = -7
    builder.font_metric.x_height = 5
    builder.font_metric.cap_height = 7
    builder.font_metric.underline_position = -1
    builder.font_metric.underline_thickness = 1
    builder.font_metric.strikeout_position = 4
    builder.font_metric.strikeout_thickness = 1

    builder.meta_info.version = '1.0.0'
    builder.meta_info.created_time = datetime.fromisoformat('2024-01-01T00:00:00Z')
    builder.meta_info.modified_time = builder.meta_info.created_time
    builder.meta_info.family_name = f'Demo {name_num}'
    builder.meta_info.weight_name = WeightName.REGULAR
    builder.meta_info.serif_style = SerifStyle.SANS_SERIF
    builder.meta_info.slant_style = SlantStyle.NORMAL
    builder.meta_info.width_style = WidthStyle.PROPORTIONAL
    builder.meta_info.manufacturer = 'Pixel Font Studio'
    builder.meta_info.designer = 'TakWolf'
    builder.meta_info.description = 'A demo font'
    builder.meta_info.copyright_info = 'Copyright (c) TakWolf'
    builder.meta_info.license_info = 'This Font Software is licensed under the SIL Open Font License, Version 1.1'
    builder.meta_info.vendor_url = 'https://github.com/TakWolf/pixel-font-builder'
    builder.meta_info.designer_url = 'https://takwolf.com'
    builder.meta_info.license_url = 'https://openfontlicense.org'

    builder.character_mapping.update(character_mapping)

    for glyph_file in glyph_files:
        horizontal_offset_x = 0
        horizontal_offset_y = (builder.font_metric.horizontal_layout.ascent + builder.font_metric.horizontal_layout.descent - glyph_file.height) // 2
        vertical_offset_x = -math.ceil(glyph_file.width / 2)
        vertical_offset_y = (builder.font_metric.font_size - glyph_file.height) // 2
        builder.glyphs.append(Glyph(
            name=_get_glyph_name(glyph_file.code_point),
            horizontal_offset=(horizontal_offset_x, horizontal_offset_y),
            advance_width=glyph_file.width,
            vertical_offset=(vertical_offset_x, vertical_offset_y),
            advance_height=builder.font_metric.font_size,
            bitmap=glyph_file.bitmap,
        ))

    builder.kerning_pairs.update(kerning_pairs)

    return builder


def main():
    outputs_dir = build_dir.joinpath('demo')
    if outputs_dir.exists():
        shutil.rmtree(outputs_dir)
    outputs_dir.mkdir(parents=True)

    character_mapping, glyph_files = _collect_glyph_files()
    for glyph_file in glyph_files:
        glyph_file.save()
    kerning_pairs = _build_kerning_pairs()

    builder = _create_builder(character_mapping, glyph_files, kerning_pairs)
    builder.save_otf(outputs_dir.joinpath('demo.otf'))
    builder.save_otf(outputs_dir.joinpath('demo.otf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_otf(outputs_dir.joinpath('demo.otf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(outputs_dir.joinpath('demo.ttf'))
    builder.save_ttf(outputs_dir.joinpath('demo.ttf.woff'), flavor=opentype.Flavor.WOFF)
    builder.save_ttf(outputs_dir.joinpath('demo.ttf.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_bdf(outputs_dir.joinpath('demo.bdf'))
    builder.save_pcf(outputs_dir.joinpath('demo.pcf'))

    collection_builder = FontCollectionBuilder()
    for index in range(100):
        builder = _create_builder(character_mapping, glyph_files, kerning_pairs, index)
        collection_builder.append(builder)
    collection_builder.save_otc(outputs_dir.joinpath('demo.otc'))
    collection_builder.save_ttc(outputs_dir.joinpath('demo.ttc'))


if __name__ == '__main__':
    main()



import math
from abc import abstractmethod
from typing import Protocol, runtime_checkable

from fontTools import cu2qu
from fontTools.misc.psCharStrings import T2CharString as OTFGlyph
from fontTools.pens.t2CharStringPen import T2CharStringPen as OTFGlyphPen
from fontTools.pens.ttGlyphPen import TTGlyphPen as TTFGlyphPen
# noinspection PyProtectedMember
from fontTools.ttLib.tables._g_l_y_f import Glyph as TTFGlyph

from pixel_font_builder.glyph import Glyph

type XTFGlyph = OTFGlyph | TTFGlyph


class OutlinesPen:
    is_ttf: bool
    pen: OTFGlyphPen | TTFGlyphPen
    current_point: tuple[float, float] | None

    def __init__(
            self,
            is_ttf: bool,
            advance_width: int,
    ):
        self.is_ttf = is_ttf
        self.pen = TTFGlyphPen() if is_ttf else OTFGlyphPen(advance_width, None)
        self.current_point = None

    def move_to(self, point: tuple[float, float]):
        self.pen.moveTo(point)
        self.current_point = point

    def line_to(self, point: tuple[float, float]):
        self.pen.lineTo(point)
        self.current_point = point

    def cubic_curve_to(
            self,
            control_point_1: tuple[float, float],
            control_point_2: tuple[float, float],
            end_point: tuple[float, float],
            double_max_err: float = 1,
    ):
        if self.is_ttf:
            self.pen.qCurveTo(*cu2qu.curve_to_quadratic([self.current_point, control_point_1, control_point_2, end_point], double_max_err)[1:])
        else:
            self.pen.curveTo(control_point_1, control_point_2, end_point)
        self.current_point = end_point

    def quadratic_curve_to(
            self,
            control_point: tuple[float, float],
            end_point: tuple[float, float],
    ):
        self.pen.qCurveTo(control_point, end_point)
        self.current_point = end_point

    def end_path(self):
        self.pen.endPath()
        self.current_point = None

    def close_path(self):
        self.pen.closePath()
        self.current_point = None

    def to_glyph(self) -> XTFGlyph:
        return self.pen.glyph() if self.is_ttf else self.pen.getCharString()


@runtime_checkable
class OutlinesPainter(Protocol):
    @abstractmethod
    def draw_outlines(self, glyph: Glyph, pen: OutlinesPen, px_to_units: int):
        raise NotImplementedError()


class SolidOutlinesPainter(OutlinesPainter):
    @staticmethod
    def _create_outlines(bitmap: list[list[int]]) -> list[list[tuple[int, int]]]:
        # 相邻像素分组
        point_groups = []
        for y, bitmap_row in enumerate(bitmap):
            for x, color in enumerate(bitmap_row):
                if color == 0:
                    continue

                new_point_group = {(x, y)}
                for point_group in reversed(point_groups):
                    # 遍历方向为右下，因此只需检查左上
                    if (x - 1, y) in point_group or (x, y - 1) in point_group:
                        point_groups.remove(point_group)
                        new_point_group.update(point_group)
                point_groups.append(new_point_group)

        # 每组生成轮廓
        outlines = []
        for point_group in point_groups:
            # 按照像素拆分线段，注意绘制顺序
            pending_segments = []
            for x, y in point_group:
                # 一个像素有左右上下四条边，如果该边没有相邻像素，则该边线段有效
                point_outline = [(x, y), (x + 1, y), (x + 1, y + 1), (x, y + 1)]
                if x <= 0 or bitmap[y][x - 1] <= 0:  # 左
                    pending_segments.append([point_outline[3], point_outline[0]])
                if x >= len(bitmap[y]) - 1 or bitmap[y][x + 1] <= 0:  # 右
                    pending_segments.append([point_outline[1], point_outline[2]])
                if y <= 0 or bitmap[y - 1][x] <= 0:  # 上
                    pending_segments.append([point_outline[0], point_outline[1]])
                if y >= len(bitmap) - 1 or bitmap[y + 1][x] <= 0:  # 下
                    pending_segments.append([point_outline[2], point_outline[3]])

            # 连接所有的线段，注意绘制顺序
            solved_segments = []
            while len(pending_segments) > 0:
                pending_segment = pending_segments.pop()
                for solved_segment in reversed(solved_segments):
                    # 一共四种连接情况
                    left_segment = None
                    right_segment = None
                    if pending_segment[-1] == solved_segment[0]:
                        left_segment = pending_segment
                        right_segment = solved_segment
                    elif pending_segment[-1] == solved_segment[-1]:
                        solved_segment.reverse()
                        left_segment = pending_segment
                        right_segment = solved_segment
                    elif solved_segment[-1] == pending_segment[0]:
                        left_segment = solved_segment
                        right_segment = pending_segment
                    elif solved_segment[-1] == pending_segment[-1]:
                        pending_segment.reverse()
                        left_segment = solved_segment
                        right_segment = pending_segment

                    # 需要连接的情况
                    if left_segment is not None and right_segment is not None:
                        solved_segments.remove(solved_segment)

                        # 连接的两个点是重复的
                        right_segment.pop(0)

                        # 判断连接的点是不是可省略
                        x, y = left_segment[-1]
                        xl, yl = left_segment[-2]
                        xr, yr = right_segment[0]
                        if (x == xl and x == xr) or (y == yl and y == yr):
                            left_segment.pop()

                        pending_segment = left_segment + right_segment
                solved_segments.append(pending_segment)

            # 将连接好的线段添加到轮廓数组中，有多条线段的情况，是中间有镂空（绘制顺序与外边框相反）
            for solved_segment in solved_segments:
                # 首尾的两个点是重复的
                solved_segment.pop(0)

                # 判断尾点是不是可省略
                x, y = solved_segment[-1]
                xl, yl = solved_segment[-2]
                xr, yr = solved_segment[0]
                if (x == xl and x == xr) or (y == yl and y == yr):
                    solved_segment.pop()

                outlines.append(solved_segment)
        return outlines

    def draw_outlines(self, glyph: Glyph, pen: OutlinesPen, px_to_units: int):
        outlines = self._create_outlines(glyph.bitmap)
        for outline in outlines:
            for index, (x, y) in enumerate(outline):
                x = (x + glyph.horizontal_offset_x) * px_to_units
                y = (glyph.height + glyph.horizontal_offset_y - y) * px_to_units

                if index == 0:
                    pen.move_to((x, y))
                else:
                    pen.line_to((x, y))
            pen.close_path()


class SquareDotOutlinesPainter(OutlinesPainter):
    size: float

    def __init__(self, size: float = 0.8):
        self.size = size

    def draw_outlines(self, glyph: Glyph, pen: OutlinesPen, px_to_units: int):
        size = self.size * px_to_units
        offset = (1 - self.size) / 2 * px_to_units
        for y, bitmap_row in enumerate(glyph.bitmap):
            y = (glyph.height + glyph.horizontal_offset_y - y) * px_to_units - offset
            for x, color in enumerate(bitmap_row):
                x = (x + glyph.horizontal_offset_x) * px_to_units + offset
                if color != 0:
                    pen.move_to((x, y))
                    pen.line_to((x + size, y))
                    pen.line_to((x + size, y - size))
                    pen.line_to((x, y - size))
                    pen.close_path()


class CircleDotOutlinesPainter(OutlinesPainter):
    radius: float

    def __init__(self, radius: float = 0.4):
        self.radius = radius

    def draw_outlines(self, glyph: Glyph, pen: OutlinesPen, px_to_units: int):
        radius = self.radius * px_to_units
        c = radius * 4 / 3 * (math.sqrt(2) - 1)
        for y, bitmap_row in enumerate(glyph.bitmap):
            y = (glyph.height + glyph.horizontal_offset_y - y - 0.5) * px_to_units
            for x, color in enumerate(bitmap_row):
                x = (x + glyph.horizontal_offset_x + 0.5) * px_to_units
                if color != 0:
                    pen.move_to((x, y + radius))
                    pen.cubic_curve_to((x + c, y + radius), (x + radius, y + c), (x + radius, y))
                    pen.cubic_curve_to((x + radius, y - c), (x + c, y - radius), (x, y - radius))
                    pen.cubic_curve_to((x - c, y - radius), (x - radius, y - c), (x - radius, y))
                    pen.cubic_curve_to((x - radius, y + c), (x - c, y + radius), (x, y + radius))
                    pen.close_path()


def create_xtf_glyphs(
        is_ttf: bool,
        outlines_painter: OutlinesPainter,
        name_to_glyph: dict[str, Glyph],
        px_to_units: int,
) -> tuple[dict[str, XTFGlyph], dict[str, tuple[int, int]], dict[str, tuple[int, int]]]:
    xtf_glyphs = {}
    horizontal_metrics = {}
    vertical_metrics = {}
    for glyph_name, glyph in name_to_glyph.items():
        advance_width = glyph.advance_width * px_to_units
        left_side_bearing = (glyph.calculate_bitmap_left_padding() + glyph.horizontal_offset_x) * px_to_units
        horizontal_metrics[glyph_name] = advance_width, left_side_bearing

        advance_height = glyph.advance_height * px_to_units
        top_side_bearing = (glyph.calculate_bitmap_top_padding() + glyph.vertical_offset_y) * px_to_units
        vertical_metrics[glyph_name] = advance_height, top_side_bearing

        pen = OutlinesPen(is_ttf, advance_width)
        outlines_painter.draw_outlines(glyph, pen, px_to_units)
        xtf_glyphs[glyph_name] = pen.to_glyph()
    return xtf_glyphs, horizontal_metrics, vertical_metrics



from __future__ import annotations

import math
import statistics
from collections import ChainMap

from pcffont import PcfFontBuilder, PcfGlyph

import pixel_font_builder
from pixel_font_builder.meta import WeightName, SlantStyle, WidthStyle

_DEFAULT_CHAR = 0xFFFE


def create_font_builder(context: pixel_font_builder.FontBuilder) -> PcfFontBuilder:
    config = context.pcf_config
    font_metric = context.font_metric
    meta_info = context.meta_info
    character_mapping = ChainMap({_DEFAULT_CHAR: '.notdef'}, context.character_mapping)
    _, name_to_glyph = context.prepare_glyphs()

    builder = PcfFontBuilder()
    builder.config.font_ascent = font_metric.horizontal_layout.ascent
    builder.config.font_descent = -font_metric.horizontal_layout.descent
    builder.config.default_char = _DEFAULT_CHAR
    builder.config.draw_right_to_left = config.draw_right_to_left
    builder.config.ms_byte_first = config.ms_byte_first
    builder.config.ms_bit_first = config.ms_bit_first
    builder.config.glyph_pad_index = config.glyph_pad_index
    builder.config.scan_unit_index = config.scan_unit_index

    for code_point, glyph_name in sorted(character_mapping.items()):
        if code_point > 0xFFFF:
            break
        glyph = name_to_glyph[glyph_name]
        builder.glyphs.append(PcfGlyph(
            name=glyph_name,
            encoding=code_point,
            scalable_width=math.ceil((glyph.advance_width / font_metric.font_size) * (75 / config.resolution_x) * 1000),
            character_width=glyph.advance_width,
            dimensions=glyph.dimensions,
            offset=glyph.horizontal_offset,
            bitmap=glyph.bitmap,
        ))

    if meta_info.manufacturer is not None:
        builder.properties.foundry = meta_info.manufacturer.replace('-', '_')
    builder.properties.family_name = meta_info.family_name.replace('-', '_')
    builder.properties.weight_name = meta_info.weight_name or WeightName.REGULAR
    if meta_info.slant_style is None or meta_info.slant_style == SlantStyle.NORMAL:
        builder.properties.slant = 'R'
    elif meta_info.slant_style == SlantStyle.ITALIC:
        builder.properties.slant = 'I'
    elif meta_info.slant_style == SlantStyle.OBLIQUE:
        builder.properties.slant = 'O'
    elif meta_info.slant_style == SlantStyle.REVERSE_ITALIC:
        builder.properties.slant = 'RI'
    elif meta_info.slant_style == SlantStyle.REVERSE_OBLIQUE:
        builder.properties.slant = 'RO'
    else:
        builder.properties.slant = 'OT'
    builder.properties.setwidth_name = 'Normal'
    if meta_info.serif_style is not None:
        builder.properties.add_style_name = meta_info.serif_style
    builder.properties.pixel_size = font_metric.font_size
    builder.properties.point_size = font_metric.font_size * 10
    builder.properties.resolution_x = config.resolution_x
    builder.properties.resolution_y = config.resolution_y
    if meta_info.width_style == WidthStyle.MONOSPACED:
        builder.properties.spacing = 'M'
    elif meta_info.width_style == WidthStyle.DUOSPACED:
        builder.properties.spacing = 'D'
    elif meta_info.width_style == WidthStyle.CHARACTER_CELL:
        builder.properties.spacing = 'C'
    elif meta_info.width_style == WidthStyle.PROPORTIONAL:
        builder.properties.spacing = 'P'
    builder.properties.average_width = round(statistics.fmean(glyph.character_width * 10 for glyph in builder.glyphs))
    builder.properties.charset_registry = 'ISO10646'
    builder.properties.charset_encoding = '1'
    builder.properties.generate_xlfd()

    if font_metric.x_height != 0:
        builder.properties.x_height = font_metric.x_height
    if font_metric.cap_height != 0:
        builder.properties.cap_height = font_metric.cap_height
    if font_metric.underline_thickness != 0:
        builder.properties.underline_position = font_metric.underline_position
        builder.properties.underline_thickness = font_metric.underline_thickness

    builder.properties.font_version = meta_info.version
    builder.properties.copyright = meta_info.copyright_info
    builder.properties['LICENSE'] = meta_info.license_info

    return builder




class Config:
    resolution_x: int
    resolution_y: int
    only_basic_plane: bool

    def __init__(
            self,
            resolution_x: int = 75,
            resolution_y: int = 75,
            only_basic_plane: bool = False,
    ):
        self.resolution_x = resolution_x
        self.resolution_y = resolution_y
        self.only_basic_plane = only_basic_plane



from pixel_font_builder.builder import FontBuilder, FontCollectionBuilder
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.meta import WeightName, SerifStyle, SlantStyle, WidthStyle, MetaInfo
from pixel_font_builder.metric import LineMetric, FontMetric



from collections import UserList
from os import PathLike

import bdffont
import fontTools.fontBuilder
import fontTools.ttLib
import pcffont

from pixel_font_builder import opentype, bdf, pcf
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.meta import MetaInfo
from pixel_font_builder.metric import FontMetric


class FontBuilder:
    font_metric: FontMetric
    meta_info: MetaInfo
    character_mapping: dict[int, str]
    glyphs: list[Glyph]
    kerning_pairs: dict[tuple[str, str], int]
    opentype_config: opentype.Config
    bdf_config: bdf.Config
    pcf_config: pcf.Config

    def __init__(self):
        self.font_metric = FontMetric()
        self.meta_info = MetaInfo()
        self.character_mapping = {}
        self.glyphs = []
        self.kerning_pairs = {}
        self.opentype_config = opentype.Config()
        self.bdf_config = bdf.Config()
        self.pcf_config = pcf.Config()

    def prepare_glyphs(self) -> tuple[list[str], dict[str, Glyph]]:
        glyph_order = ['.notdef']
        name_to_glyph = {}

        for glyph in self.glyphs:
            if glyph.name in name_to_glyph:
                raise RuntimeError(f'duplicate glyphs: {repr(glyph.name)}')
            if glyph.name != '.notdef':
                glyph_order.append(glyph.name)
            name_to_glyph[glyph.name] = glyph

        if '.notdef' not in name_to_glyph:
            raise RuntimeError("missing glyph: '.notdef'")

        for code_point, glyph_name in self.character_mapping.items():
            if code_point < 0:
                raise RuntimeError('code points must >= 0')
            if glyph_name not in name_to_glyph:
                raise RuntimeError(f'missing glyph: {repr(glyph_name)}')

        for (left_glyph_name, right_glyph_name), _ in self.kerning_pairs.items():
            if left_glyph_name not in name_to_glyph:
                raise RuntimeError(f'missing glyph: {repr(left_glyph_name)}')
            if right_glyph_name not in name_to_glyph:
                raise RuntimeError(f'missing glyph: {repr(right_glyph_name)}')

        return glyph_order, name_to_glyph

    def to_otf_builder(self, flavor: opentype.Flavor | None = None) -> fontTools.fontBuilder.FontBuilder:
        return opentype.create_font_builder(self, False, flavor)

    def save_otf(
            self,
            file_path: str | PathLike[str],
            flavor: opentype.Flavor | None = None,
    ):
        self.to_otf_builder(flavor).save(file_path)

    def to_ttf_builder(self, flavor: opentype.Flavor | None = None) -> fontTools.fontBuilder.FontBuilder:
        return opentype.create_font_builder(self, True, flavor)

    def save_ttf(
            self,
            file_path: str | PathLike[str],
            flavor: opentype.Flavor | None = None,
    ):
        self.to_ttf_builder(flavor).save(file_path)

    def to_bdf_builder(self) -> bdffont.BdfFont:
        return bdf.create_font_builder(self)

    def save_bdf(self, file_path: str | PathLike[str]):
        self.to_bdf_builder().save(file_path)

    def to_pcf_builder(self) -> pcffont.PcfFontBuilder:
        return pcf.create_font_builder(self)

    def save_pcf(self, file_path: str | PathLike[str]):
        self.to_pcf_builder().save(file_path)


class FontCollectionBuilder(UserList[FontBuilder]):
    def to_otc_builder(self) -> fontTools.ttLib.TTCollection:
        return opentype.create_font_collection_builder(self, False)

    def save_otc(
            self,
            file_path: str | PathLike[str],
            share_tables: bool = True,
    ):
        self.to_otc_builder().save(file_path, share_tables)

    def to_ttc_builder(self) -> fontTools.ttLib.TTCollection:
        return opentype.create_font_collection_builder(self, True)

    def save_ttc(
            self,
            file_path: str | PathLike[str],
            share_tables: bool = True,
    ):
        self.to_ttc_builder().save(file_path, share_tables)



from __future__ import annotations

from typing import Any


class LineMetric:
    ascent: int
    descent: int
    line_gap: int

    def __init__(
            self,
            ascent: int = 0,
            descent: int = 0,
            line_gap: int = 0,
    ):
        self.ascent = ascent
        self.descent = descent
        self.line_gap = line_gap

    def __mul__(self, other: Any) -> LineMetric:
        if not isinstance(other, int):
            raise TypeError(f"can't multiply 'LineMetric' by non-int of type '{type(other).__name__}'")
        return LineMetric(
            self.ascent * other,
            self.descent * other,
            self.line_gap * other,
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, LineMetric):
            return NotImplemented
        return (self.ascent == other.ascent and
                self.descent == other.descent and
                self.line_gap == other.line_gap)

    @property
    def line_height(self) -> int:
        return self.ascent - self.descent


class FontMetric:
    font_size: int
    horizontal_layout: LineMetric
    vertical_layout: LineMetric
    x_height: int
    cap_height: int
    underline_position: int
    underline_thickness: int
    strikeout_position: int
    strikeout_thickness: int

    def __init__(
            self,
            font_size: int = 0,
            horizontal_layout: LineMetric | None = None,
            vertical_layout: LineMetric | None = None,
            x_height: int = 0,
            cap_height: int = 0,
            underline_position: int = 0,
            underline_thickness: int = 0,
            strikeout_position: int = 0,
            strikeout_thickness: int = 0,
    ):
        self.font_size = font_size
        self.horizontal_layout = LineMetric() if horizontal_layout is None else horizontal_layout
        self.vertical_layout = LineMetric() if vertical_layout is None else vertical_layout
        self.x_height = x_height
        self.cap_height = cap_height
        self.underline_position = underline_position
        self.underline_thickness = underline_thickness
        self.strikeout_position = strikeout_position
        self.strikeout_thickness = strikeout_thickness

    def __mul__(self, other: Any) -> FontMetric:
        if not isinstance(other, int):
            raise TypeError(f"can't multiply 'FontMetric' by non-int of type '{type(other).__name__}'")
        return FontMetric(
            self.font_size * other,
            self.horizontal_layout * other,
            self.vertical_layout * other,
            self.x_height * other,
            self.cap_height * other,
            self.underline_position * other,
            self.underline_thickness * other,
            self.strikeout_position * other,
            self.strikeout_thickness * other,
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FontMetric):
            return NotImplemented
        return (self.font_size == other.font_size and
                self.horizontal_layout == other.horizontal_layout and
                self.vertical_layout == other.vertical_layout and
                self.x_height == other.x_height and
                self.cap_height == other.cap_height and
                self.underline_position == other.underline_position and
                self.underline_thickness == other.underline_thickness and
                self.strikeout_position == other.strikeout_position and
                self.strikeout_thickness == other.strikeout_thickness)
